import numpy as np

try:
    import ipywidgets
except ImportError:
    ipywidgets = None

try:
    import matplotlib.pyplot as plt
    from mpl_toolkits.mplot3d import Axes3D  # Activates 3d perspective
    from matplotlib import cm
except ImportError:
    plt = None

from .variables import FuzzyVariable
from .rules import FuzzyRuleSet


class FIS:
    """A fuzzy inference system"""

    def __init__(self, variables, rules, target):
        self.variables = variables
        self.rules = rules

        # TODO: Support multitarget
        self.target = target

    def get_description(self):
        return {"variables": [v.get_description() for v in self.variables],
                "rules": self.rules.get_description(),
                "target": self.target.get_description()}

    @staticmethod
    def from_description(description):
        variables = [FuzzyVariable.from_description(d) for d in description["variables"]]
        target_variable = FuzzyVariable.from_description(description["target"])
        variables_dict = {**{v.name: v for v in variables}, target_variable.name: target_variable}
        return FIS(variables,
                   FuzzyRuleSet.from_description(description["rules"], variables_dict),
                   target_variable)

    def get_output(self, values):
        return self.rules(values)

    def get_crisp_output(self, values):
        return self.target.domain.centroid(self.get_output(values))

    def get_interactive(self):
        if ipywidgets is None or plt is None:
            raise ModuleNotFoundError("ipywidgets and matplotlib are required")

        def plot(**kwargs):
            output = self.get_output(kwargs)
            self.target.domain.plot_set(output)
            plt.vlines(self.target.domain.centroid(output), *plt.ylim(), color="red")
            plt.legend(["Fuzzy output", "Centroid"])
            plt.show()

        ipywidgets.interact(plot, **{variable.name: variable.domain.get_ipywidget() for variable in self.variables})

    def plot_surface(self):
        if len(self.variables) < 2:
            raise ValueError("At least two variables are required")
        if len(self.variables) > 2:
            # TODO: Allow selection of additional variables as fixed values
            raise NotImplementedError

        x_name = self.variables[0].name
        y_name = self.variables[1].name
        # TODO: Allow coarser mesh
        xx = self.variables[0].domain.get_mesh()
        yy = self.variables[1].domain.get_mesh()
        zz = np.asarray(
            [
                [
                    self.get_crisp_output({x_name: x, y_name: y})
                    for x in xx
                ]
                for y in yy
            ]
        )

        fig = plt.figure()
        ax = fig.add_subplot(111, projection="3d")

        ax.plot_surface(*np.meshgrid(xx, yy), zz, cmap=cm.viridis)

        ax.set_xlabel(x_name)
        ax.set_ylabel(y_name)
        ax.set_zlabel(self.target.name)

        ax.invert_xaxis()  # Seems more natural to me

        return ax



import numpy as np

try:
    import matplotlib.pyplot as plt
except ImportError:
    plt = None

try:
    import ipywidgets
except ImportError:
    ipywidgets = None


class Domain:
    """A domain where diffuse sets are defined"""

    def __init__(self, name):
        self.name = name

    @staticmethod
    def from_description(description):
        return _domain_subclasses[description["type"]].from_description(description)

    def get_mesh(self):
        """Get a mesh representing the domain"""
        raise NotImplementedError("A domain subclass must be used instead.")

    def evaluate_set(self, set):
        """Get a pair of list with a mesh representing the domain and the evaluation of a membership function on it"""
        mesh = self.get_mesh()
        return mesh, [set(x) for x in mesh]

    def plot_set(self, set, **kwargs):
        """Plot a fuzzy set"""
        if plt is None:
            raise ModuleNotFoundError("Matplotlib is required for plotting")
        plt.plot(*self.evaluate_set(set), **kwargs)
        plt.xlabel(self.name)
        plt.ylabel("Membership function")

    def centroid(self, set):
        """Calculate the centroid of a fuzzy set"""
        raise NotImplementedError("A Domain subclass must be used instead.")

    def get_ipywidget(self):
        """Get a widget representing the domain"""
        raise NotImplementedError("A Domain subclass must be used instead.")


class FloatDomain(Domain):
    def __init__(self, name, min, max, steps):
        """

        Args:
            name (str): Name of the domain.
            min (float): Minimum value of the domain
            max (float): Maximum value of the domain
            steps (int or float): Number of steps if int or step size if float.
        """
        super().__init__(name)
        self.min = min
        self.max = max
        self.steps = steps

    def get_description(self):
        return {"type": "FloatDomain", "name": self.name, "min": self.min, "max": self.max, "steps": self.steps}

    @staticmethod
    def from_description(description):
        return FloatDomain(description["name"], description["min"], description["max"], description["steps"])

    def get_mesh(self):
        if self.steps is None or isinstance(self.steps, int):
            return np.linspace(self.min, self.max, self.steps)
        elif isinstance(self.steps, float):
            return np.arange(self.min, self.max, self.steps)
        else:
            raise ValueError("Bad type for steps")

    def centroid(self, set):
        xx, mu = self.evaluate_set(set)
        return np.average(xx, weights=mu)

    def get_ipywidget(self):
        if ipywidgets is None:
            raise ModuleNotFoundError("ipywidgets is required")
        return ipywidgets.FloatSlider(min=self.min, max=self.max)


_domain_subclasses = {"FloatDomain": FloatDomain}



from math import exp


class FuzzySet:
    """A fuzzy set"""

    def __init__(self, mu=None):
        self.mu = mu

    def get_description(self):
        raise NotImplementedError("Descriptions are only available for primitive types")

    @staticmethod
    def from_description(description):
        return _set_types[description["type"]].from_description(description)

    def __call__(self, x):
        assert self.mu is not None, "A membership function has to be defined"
        return self.mu(x)

    # TODO: More generalized fuzzy set operations could be defined

    def __neg__(self):
        return FuzzySet(lambda x: 1 - self(x))

    def __or__(self, other):
        return FuzzySet(lambda x: max(self(x), other(x)))

    def __and__(self, other):
        return FuzzySet(lambda x: min(self(x), other(x)))

    @staticmethod
    def n_ary_or(sets):
        return FuzzySet(lambda x: max(s(x) for s in sets))

    @staticmethod
    def n_ary_and(sets):
        return FuzzySet(lambda x: min(s(x) for s in sets))


class SingletonSet(FuzzySet):
    """A singleton fuzzy set (Kronecker delta)"""

    def __init__(self, x):
        """

        Args:
            x: The unique value where membership is 1.
        """
        super().__init__(lambda y: 1 if x == y else 0)
        self.x = x

    def get_description(self):
        return {"type": "singleton", "x": self.x}

    @staticmethod
    def from_description(description):
        return SingletonSet(description["x"])


class DiscreteFuzzySet(FuzzySet):
    """A discrete fuzzy set (non-null in a discrete set of points)"""

    def __init__(self, d):
        """

        Args:
            d (dict): Mapping of values to non-null membership values

        """
        self.d = d
        super().__init__()

    def __call__(self, x):
        return self.d.get(x, 0)

    def get_description(self):
        return {"type": "discrete", "d": self.d}

    @staticmethod
    def from_description(description):
        return DiscreteFuzzySet(description["d"])


class TriangularFuzzySet(FuzzySet):
    """A fuzzy set defined by a triangular function"""

    def __init__(self, a, b, c):
        self.a = a
        self.b = b
        self.c = c
        super().__init__()

    def get_description(self):
        return {"type": "triangular", "a": self.a, "b": self.b, "c": self.c}

    @staticmethod
    def from_description(description):
        return TriangularFuzzySet(description["a"], description["b"], description["c"])

    def __call__(self, x):
        return max(min((x - self.a) / (self.b - self.a), (self.c - x) / (self.c - self.b)), 0)


class TrapezoidalFuzzySet(FuzzySet):
    """A fuzzy set defined by a trapezoidal function"""

    def __init__(self, a, b, c, d):
        self.a = a
        self.b = b
        self.c = c
        self.d = d
        super().__init__()

    def get_description(self):
        return {"type": "trapezoidal", "a": self.a, "b": self.b, "c": self.c, "d": self.d}

    @staticmethod
    def from_description(description):
        return TrapezoidalFuzzySet(description["a"], description["b"], description["c"], description["d"])

    def __call__(self, x):
        return max(min((x - self.a) / (self.b - self.a), 1, (self.d - x) / (self.d - self.c)), 0)


class GaussianFuzzySet(FuzzySet):
    """A fuzzy set defined by a gaussian function"""

    def __init__(self, a, c):
        self.a = a
        self.c = c
        super().__init__()

    def get_description(self):
        return {"type": "gaussian", "a": self.a, "c": self.c}

    @staticmethod
    def from_description(description):
        return GaussianFuzzySet(description["a"], description["c"])

    def __call__(self, x):
        return exp(-((x - self.c) / self.a) ** 2 / 2)


class BellFuzzySet(FuzzySet):
    """A fuzzy set defined by a generalized Bell MF

    :math:`\\mu_{a,b,c}(x)= \\frac{1}{1+\\left|\frac{x-c}{a}\\right|^{2b}}`
    """

    def __init__(self, a, b, c):
        self.a = a
        self.b = b
        self.c = c
        super().__init__()

    def get_description(self):
        return {"type": "bell", "a": self.a, "b": self.b, "c": self.c}

    @staticmethod
    def from_description(description):
        return BellFuzzySet(description["a"], description["b"], description["c"])

    def __call__(self, x):
        return 1 / (1 + ((x - self.c) / self.a) ** (2 * self.b))


class SigmoidalFuzzySet(FuzzySet):
    """A fuzzy set defined by a sigmoid function"""

    def __init__(self, a, c):
        self.a = a
        self.c = c
        super().__init__()

    def get_description(self):
        return {"type": "sigmoidal", "a": self.a, "c": self.c}

    @staticmethod
    def from_description(description):
        return SigmoidalFuzzySet(description["a"], description["c"])

    def __call__(self, x):
        return 1 / (1 + exp(-self.a * (x - self.c)))


_set_types = {"singleton": SingletonSet, "discrete": DiscreteFuzzySet, "sigmoid": SigmoidalFuzzySet,
              "bell": BellFuzzySet, "gaussian": GaussianFuzzySet, "trapezoidal": TrapezoidalFuzzySet,
              "triangular": TriangularFuzzySet}



"""zadeh - Python package to build fuzzy inference systems"""

__version__ = '0.1.0'
__author__ = 'Dih5 <dihedralfive@gmail.com>'
__all__ = []

from .domains import *
from .sets import *
from .variables import *
from .rules import *
from .fis import *



from .sets import FuzzySet


class FuzzyProposition:
    """A fuzzy-logic proposition"""

    def __init__(self):
        pass

    def __call__(self, values):
        """Evaluate the statement, returning a number in [0, 1]"""
        raise NotImplementedError

    @staticmethod
    def from_description(description, variables_dict):
        return _fuzzy_propositions[description["type"]].from_description(description, variables_dict)

    def __repr__(self):
        return "FuzzyProposition<%s>" % str(self)

    def __str__(self):
        return "[Undefined proposition]"

    # TODO: More generalized fuzzy set operations could be defined

    def __or__(self, other):
        return FuzzyOr([self, other])

    def __and__(self, other):
        return FuzzyAnd([self, other])

    def __invert__(self):
        return FuzzyNot(self)


class FuzzyValuation(FuzzyProposition):
    """An elemental fuzzy proposition of the form '<variable> is <value>'"""

    def __init__(self, variable, value):
        super().__init__()
        self.variable = variable
        self.value = value

    def get_description(self):
        return {"type": "is", "variable": self.variable.name, "value": self.value}

    @staticmethod
    def from_description(description, variables_dict):
        return FuzzyValuation(variables_dict[description["variable"]], description["value"])

    def __call__(self, values):
        return self.variable[self.value](values[self.variable.name])

    def __str__(self):
        return "%s is %s" % (self.variable.name, self.value)


class FuzzyNotValuation(FuzzyProposition):
    """An 'elemental' fuzzy proposition of the form <variable> is not <value>

    While this could be described using the unitary negation operator, the class is provided for convenience

    """

    def __init__(self, variable, value):
        super().__init__()
        self.variable = variable
        self.value = value

    def get_description(self):
        return {"type": "is not", "variable": self.variable.name, "value": self.value}

    @staticmethod
    def from_description(description, variables_dict):
        return FuzzyNotValuation(variables_dict[description["variable"]], description["value"])

    def __call__(self, values):
        return 1 - self.variable[self.value](values[self.variable.name])

    def __str__(self):
        return "%s is not %s" % (self.variable.name, self.value)


class FuzzyNot(FuzzyProposition):
    """A fuzzy proposition of the form 'not <p>'"""

    def __init__(self, proposition):
        super().__init__()
        self.proposition = proposition

    def get_description(self):
        return {"type": "not", "children": [self.proposition.get_description()]}

    @staticmethod
    def from_description(description, variables_dict):
        return FuzzyNot(variables_dict[description["children"][0]])

    def __call__(self, values):
        return 1 - self.proposition(values)

    def __str__(self):
        return "not (%s)" % str(self.proposition)


class FuzzyAnd(FuzzyProposition):
    """A fuzzy proposition of the form <p1> and <p2>"""

    def __init__(self, proposition_list):
        super().__init__()
        self.proposition_list = proposition_list

    def get_description(self):
        return {"type": "and", "children": [p.get_description() for p in self.proposition_list]}

    @staticmethod
    def from_description(description, variables_dict):
        return FuzzyAnd([variables_dict[variable] for variable in description["children"]])

    def __call__(self, values):
        return min(p(values) for p in self.proposition_list)

    def __str__(self):
        return " and ".join("(%s)" % str(p) for p in self.proposition_list)


class FuzzyOr(FuzzyProposition):
    """A fuzzy proposition of the form <p1> or <p2>"""

    def __init__(self, proposition_list):
        super().__init__()
        self.proposition_list = proposition_list

    def get_description(self):
        return {"type": "or", "children": [p.get_description() for p in self.proposition_list]}

    @staticmethod
    def from_description(description, variables_dict):
        return FuzzyOr([FuzzyProposition.from_description(variable, variables_dict) for variable in description["children"]])

    def __call__(self, values):
        return max(p(values) for p in self.proposition_list)

    def __str__(self):
        return " or ".join("(%s)" % str(p) for p in self.proposition_list)


_fuzzy_propositions = {"or": FuzzyOr, "and": FuzzyAnd, "not": FuzzyNot, "is": FuzzyValuation,
                       "is not": FuzzyNotValuation}


class FuzzyRule:
    """A fuzzy rule of the form 'if <antecedent> then <consequent>'"""

    def __init__(self, antecedent, consequent):
        super().__init__()
        self.antecedent = antecedent
        self.consequent = consequent

        if not isinstance(consequent, FuzzyValuation):
            # TODO: Support this
            raise ValueError("Complex consequent rules not supported")

    def get_description(self):
        return {"antecedent": self.antecedent.get_description(), "consequent": self.consequent.get_description()}

    @staticmethod
    def from_description(description, variables_dict):
        return FuzzyRule(FuzzyProposition.from_description(description["antecedent"], variables_dict),
                         FuzzyProposition.from_description(description["consequent"], variables_dict))

    def __call__(self, values):
        """Evaluate the rule, returning a fuzzy number"""
        # TODO: Add more methods
        # Mamdani inference
        cutoff = self.antecedent(values)
        return FuzzySet(lambda x: min(cutoff, self.consequent.variable[self.consequent.value](x)))

    def __repr__(self):
        return "FuzzyRule<%s>" % str(self)

    def __str__(self):
        return "if (%s) then (%s)" % (self.antecedent, self.consequent)


class FuzzyRuleSet:
    """A set of fuzzy rules"""

    def __init__(self, rule_list):
        super().__init__()
        self.rule_list = rule_list


    def get_description(self):
        return {"rule_list": [r.get_description() for r in self.rule_list]}

    @staticmethod
    def from_description(description, variables_dict):
        return FuzzyRuleSet([FuzzyRule.from_description(d, variables_dict) for d in description["rule_list"]])



    def __call__(self, values):
        """Evaluate the set of rules, returning a fuzzy number"""
        # TODO: Consider multiple output
        # TODO: Add more methods
        # Mamdani inference
        return FuzzySet.n_ary_or([rule(values) for rule in self.rule_list])

    def __getitem__(self, item):
        return self.rule_list[item]

    def __repr__(self):
        return "FuzzyRuleSet<%s>" % str(self)

    def __str__(self):
        return "\n".join(str(s) for s in self.rule_list)

    def __iter__(self):
        return iter(self.rule_list)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script"""

import os
from setuptools import setup, find_packages

# Get the long description from the README file
with open(os.path.join(os.path.abspath(os.path.dirname(__file__)), 'README.md')) as f:
    long_description = f.read()

setup(author="Dih5",
      author_email='dihedralfive@gmail.com',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Intended Audience :: Science/Research',
          'Natural Language :: English',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
      ],
      description='Python package to build fuzzy inference systems',
      extras_require={
          "docs": ["nbsphinx", "sphinx-rtd-theme", "IPython"],
          "test": ["pytest"],
          "extras": ["matplotlib", "ipywidgets"],
      },
      keywords=[],
      long_description=long_description,
      long_description_content_type='text/markdown',
      name='zadeh',
      packages=find_packages(include=['zadeh'], exclude=["demos", "tests", "docs"]),
      install_requires=["numpy"],
      url='https://github.com/dih5/zadeh',
      version='0.1.0',

      )



try:
    import matplotlib.pyplot as plt
except ImportError:
    plt = None

from .domains import Domain
from .sets import FuzzySet


class FuzzyVariable:
    """A fuzzy variable"""

    def __init__(self, domain, values, name=None):
        """

        Args:
            domain (Domain): Domain (universe of discourse) of the variable.
            values (dict of str to FuzzySet): Mapping of values to fuzzy numbers.
            name (str): The name of the variable. If none, taken from the domain name.
        """
        self.domain = domain
        self.values = values
        self.name = name if name is not None else self.domain.name

    def get_description(self):
        return {"name": self.name,
                "values": {key: val.get_description() for key, val in self.values.items()},
                "domain": self.domain.get_description()}

    @staticmethod
    def from_description(description):
        domain = Domain.from_description(description["domain"])
        values = {key: FuzzySet.from_description(val) for key, val in description["values"].items()}
        return FuzzyVariable(domain, values, name=description["name"])

    def plot(self):
        """Plot the membership function for each of the values"""
        for value, set in self.values.items():
            self.domain.plot_set(set, label=value)

        plt.legend()

    def __getitem__(self, item):
        return self.values[item]
